﻿ /*
| Tracker Controller by AlienAxeMan (3/7/2023)
| https://www.patreon.com/AlienAxeMan
 \
  | This is a script intended for use within Virt-A-Mate, very handy as a session plugin.
  | Upon scene load, it will collect any people into lists by gender and start a keyword
  | recognizer with a set of keywords that controlls various links from objects to
  | trackers.
  |
  | This voice recognition system relies exclusively on the system default microphone!
  |
  | Licensed under CC BY (see https://creativecommons.org/licenses/by/4.0/)
*/

/* TODO: expose more configuration in the plugin UI, such as:
 *  - add a toggle for enabling/disabling speech recognition (global - hiding/disabling each keyword configuration, as well as local - per keyword)
*/ 

using System;
using UnityEngine;
using System.Collections.Generic;
using System.Text;
using UnityEngine.Windows.Speech;
using static FreeControllerV3;
using UnityEngine.Experimental.UIElements;
using UnityEngine.UI;
using Leap.Unity.Query;
using System.Linq;
using static SuperController;

namespace AlienAxeMan
{
    internal class AAM_TrackerController : MVRScript
    {
        private string PLUGIN_LABEL = "Tracker Controller";
        private string PLUGIN_VERSION = "1.0d";

        private KeywordRecognizer m_recognizer = null;
        private List<string> m_keywords_internal = null;
        private Dictionary<string, List<Action>> m_keyword_to_action_map_internal;

        private JSONStorableBool m_log_keyword_recognition;

        List<Atom> m_male_atoms = new List<Atom>();
        List<Atom> m_female_atoms = new List<Atom>();

        List<string> m_tracker_ids;

        List<TrackerMapper> m_tracker_mappers = new List<TrackerMapper>();

        private class TrackerMapper
        {
            private AAM_TrackerController m_tracker_controller = null;

            private readonly List<string> m_empty_strings_list = new List<string>();
            private readonly List<string> m_automatic_choices = new List<string>() { "<Auto> Dildo", "<Auto> Female", "<Auto> Male" };

            private string m_id;

            JSONStorableStringChooser m_tracker_id_chooser;
            JSONStorableString m_start_keyword;
            JSONStorableString m_stop_keyword;

            public JSONStorableStringChooser m_atom_chooser;
            JSONStorableStringChooser m_rigid_body_chooser;

            public void StartTextChanged(string text)
            {
                //SuperController.LogMessage("TrackerMapper" + m_id + " start text changed to: " + text);

                m_start_keyword.val = text;
                m_tracker_controller.RefreshKeywords();
            }

            public void StopTextChanged(string text)
            {
                //SuperController.LogMessage("TrackerMapper" + m_id + " stop text changed to: " + text);

                m_stop_keyword.val = text;
                m_tracker_controller.RefreshKeywords();
            }

            public Atom GetAtomByUID(string uid)
            {
                //SuperController.LogMessage("GetAtomByUID: " + uid);
                switch (uid)
                {
                    case "<Auto> Dildo":
                        {
                            //SuperController.LogMessage("<Auto> Dildo>...");
                            Atom dildo = SuperController.singleton.GetAtomByUid("Dildo");
                            if (dildo != null)
                            {
                                //SuperController.LogMessage("Found <Auto> Dildo>");
                                return dildo;
                            }
                            break;
                        }
                    case "<Auto> Female":
                        {
                            //SuperController.LogMessage("<Auto> Female count: " + m_tracker_controller.m_female_atoms.Count.ToString());
                            if (m_tracker_controller.m_female_atoms.Count > 0)
                            {
                                //SuperController.LogMessage("Found <Auto> Female");
                                return m_tracker_controller.m_female_atoms[0];
                            }
                            break;
                        }
                    case "<Auto> Male":
                        {
                            //SuperController.LogMessage("<Auto> Male count: " + m_tracker_controller.m_male_atoms.Count.ToString());
                            if (m_tracker_controller.m_male_atoms.Count > 0)
                            {
                                //SuperController.LogMessage("Found <Auto> Male");
                                return m_tracker_controller.m_male_atoms[0];
                            }
                            break;
                        }
                    default:
                        {
                            //SuperController.LogMessage("\tRerouting to SuperController.singleton.GetAtomByUid");
                            return SuperController.singleton.GetAtomByUid(uid);
                        }
                }

                return null;
            }

            public void SyncDropDowns(string val)
            {
                //SuperController.LogMessage("PersonAtomSetChoice..." + val);
                m_atom_chooser.choices = m_automatic_choices.Concat(SuperController.singleton.GetAtomUIDs()).ToList();
                //foreach (var choice in m_atom_chooser.choices)
                //{
                //    SuperController.LogMessage("\t" + choice);
                //}
                var atom = m_atom_chooser.val != null ? GetAtomByUID(m_atom_chooser.val) : null;
                if (atom == null)
                {
                    m_rigid_body_chooser.choices = m_empty_strings_list;
                    //if (m_rigid_body_chooser.val != null)
                    //{
                    //    m_rigid_body_chooser.val = null;
                    //}
                }
                else
                {
                    m_rigid_body_chooser.choices = atom.linkableRigidbodies.Select(rb => rb.name).ToList();
                    if (!m_rigid_body_chooser.choices.Contains(m_rigid_body_chooser.val))
                    {
                        m_rigid_body_chooser.val = m_rigid_body_chooser.choices.FirstOrDefault();
                    }
                }
            }

            public void RefreshAtoms()
            {
                SyncDropDowns(m_atom_chooser.val);
            }

            public void AddKeywords()
            {
                m_tracker_controller.AddInternalKeywordAction(m_start_keyword.val, StartControlling);
                m_tracker_controller.AddInternalKeywordAction(m_stop_keyword.val, StopControlling);
            }

            public TrackerMapper(AAM_TrackerController tracker_controller, string label, string default_atom_uid = null, string default_rigid_body_uid = null, string default_tracker_id = null)
            {
                // TODO: long term, add support for an optional auto-selector, such as "1st Female headControl" (using fixed sub-element chooser choices)
                // this should work as a scene plugin, across all scenes, etc.

                // Create Atom/Object picker (ex. "Woman")
                // Create Sub-Element picker (ex. "headControl")

                m_tracker_controller = tracker_controller;
                m_id = m_tracker_controller.m_tracker_mappers.Count.ToString();

                m_atom_chooser = new JSONStorableStringChooser("Atom" + m_id, m_empty_strings_list, default_atom_uid, "Atom", SyncDropDowns);
                m_tracker_controller.RegisterStringChooser(m_atom_chooser);
                m_tracker_controller.CreateScrollablePopup(m_atom_chooser, false);//.popupPanelHeight = 900;

                m_rigid_body_chooser = new JSONStorableStringChooser("Rigidbody" + m_id, m_empty_strings_list, default_rigid_body_uid, "Rigidbody");
                m_tracker_controller.RegisterStringChooser(m_rigid_body_chooser);
                m_tracker_controller.CreateScrollablePopup(m_rigid_body_chooser, true);//.popupPanelHeight = 800;

                m_tracker_id_chooser = m_tracker_controller.CreateTrackerIDChooser("Tracker" + m_id, "Tracker", default_tracker_id);

                m_tracker_controller.CreateSpacer(false).height = 25;

                m_start_keyword = new JSONStorableString("StartKeyword" + m_id, "Control " + label);
                m_tracker_controller.RegisterString(m_start_keyword);
                m_start_keyword.setCallbackFunction = StartTextChanged;
                UIDynamicTextField m_dildo_start_keyword_text_field = m_tracker_controller.CreateTextField(m_start_keyword, true);
                m_dildo_start_keyword_text_field.backgroundColor = Color.white;
                LayoutElement layout = m_dildo_start_keyword_text_field.GetComponent<LayoutElement>();
                layout.preferredHeight = layout.minHeight = 35;
                InputField inputfield = m_dildo_start_keyword_text_field.gameObject.AddComponent<InputField>();
                inputfield.textComponent = m_dildo_start_keyword_text_field.UItext;
                inputfield.text = m_start_keyword.val;
                m_start_keyword.RegisterInputField(inputfield);

                m_stop_keyword = new JSONStorableString("StopKeyword" + m_id, "Stop Controlling " + label);
                m_tracker_controller.RegisterString(m_stop_keyword);
                m_stop_keyword.setCallbackFunction = StopTextChanged;
                UIDynamicTextField m_dildo_stop_keyword_text_field = m_tracker_controller.CreateTextField(m_stop_keyword, true);
                m_dildo_stop_keyword_text_field.backgroundColor = Color.white;
                layout = m_dildo_stop_keyword_text_field.GetComponent<LayoutElement>();
                layout.preferredHeight = layout.minHeight = 35;
                inputfield = m_dildo_stop_keyword_text_field.gameObject.AddComponent<InputField>();
                inputfield.textComponent = m_dildo_stop_keyword_text_field.UItext;
                inputfield.text = m_stop_keyword.val;
                m_stop_keyword.RegisterInputField(inputfield);

                m_tracker_controller.CreateSpacer(true).height = 20;

                AddKeywords();
            }

            public void StartControlling()
            {
                Atom atom = GetAtomByUID(m_atom_chooser.val);
                if (atom == null)
                {
                    return;
                }

                FreeControllerV3 controller = atom.GetStorableByID(m_rigid_body_chooser.val) as FreeControllerV3;
                if (controller == null)
                {
                    SuperController.LogError("StartControlling: Failed to get controller for " + m_rigid_body_chooser.val.ToString());
                    return;
                }

                Rigidbody tracker_rigid_body = SuperController.singleton.RigidbodyNameToRigidbody("[CameraRig]:" + m_tracker_id_chooser.val);
                if (tracker_rigid_body == null)
                {
                    SuperController.LogError("StartControlling: Failed to get rigid body for [CameraRig]:" + m_tracker_id_chooser.val.ToString());
                    return;
                }

                SuperController.LogMessage("StartControlling \"" + m_atom_chooser.val.ToString() + ":" + m_rigid_body_chooser.val + "\" by Rigidbody: " + tracker_rigid_body.ToString());
                controller.SelectLinkToRigidbody(tracker_rigid_body, SelectLinkState.PositionAndRotation, true, true);
            }

            public void StopControlling()
            {
                Atom atom = GetAtomByUID(m_atom_chooser.val);
                if (atom == null)
                {
                    return;
                }

                FreeControllerV3 controller = atom.GetStorableByID(m_rigid_body_chooser.val) as FreeControllerV3;
                if (controller == null)
                {
                    return;
                }

                controller.SelectLinkToRigidbody(null, SelectLinkState.PositionAndRotation, false, true);
            }
        }

        public void RefreshKeywords()
        {
            if (m_recognizer != null)
            {
                m_recognizer.Stop();
                m_recognizer.Dispose();
            }

            if (m_keywords_internal == null)
            {
                m_keywords_internal = new List<string>();
            }
            else
            {
                m_keywords_internal.Clear();
            }

            if (m_keyword_to_action_map_internal != null)
            {
                m_keyword_to_action_map_internal.Clear();
            }

            foreach (TrackerMapper tracker_mapper in m_tracker_mappers)
            {
                tracker_mapper.AddKeywords();
            }

            //SuperController.LogMessage("Internal Keywords...");
            //foreach (string keyword in m_keywords_internal)
            //{
            //    SuperController.LogMessage("\t" + keyword);
            //}

            if (m_keywords_internal.Count > 0)
            {
                m_recognizer = new KeywordRecognizer(m_keywords_internal.ToArray());
                m_recognizer.OnPhraseRecognized += OnPhraseRecognizedInternal;
                m_recognizer.Start();
            }
        }

        public void CollectAtomsByGender()
        {
            m_male_atoms.Clear();
            m_female_atoms.Clear();

            List<Atom> atoms = GetSceneAtoms();
            foreach (var atom in atoms)
            {
                DAZCharacterSelector selector = atom.GetComponentInChildren<DAZCharacterSelector>();
                if (selector != null)
                {
                    if (selector.gender == DAZCharacterSelector.Gender.Female)
                    {
                        //SuperController.LogMessage(atom.ToString() + " is a Female");
                        m_female_atoms.Add(atom);
                    }

                    if (selector.gender == DAZCharacterSelector.Gender.Male)
                    {
                        //SuperController.LogMessage(atom.ToString() + " is a Male");
                        m_male_atoms.Add(atom);
                    }
                }
            }
        }

        private void OnPhraseRecognizedInternal(PhraseRecognizedEventArgs args)
        {
            if (m_log_keyword_recognition.val)
            {
                StringBuilder builder = new StringBuilder();
                builder.AppendFormat("Speech Recognized: {0} ({1})", args.text, args.confidence);
                SuperController.LogMessage(builder.ToString());
            }

            List<Action> actions = null;
            if (m_keyword_to_action_map_internal.TryGetValue(args.text, out actions))
            {
                foreach (var action in actions)
                {
                    action.Invoke();
                }
            }
            else
            {
                SuperController.LogError("Failed to get action(s) for " + args.text);
                foreach (KeyValuePair<string, List<Action>> entry in m_keyword_to_action_map_internal)
                {
                    SuperController.LogError(entry.Key + ": " + entry.Value.ToString());
                }
            }
        }

        public void OnAtomUIDsChanged(List<string> atom_uids)
        {
            //SuperController.LogMessage("AtomUIDsChanged");
            CollectAtomsByGender();

            foreach (var tracker in m_tracker_mappers)
            {
                tracker.RefreshAtoms();
            }
        }

        public void OnEnable()
        {
            //SuperController.LogMessage(nameof(AAM_TrackerController) + ":" + "OnEnable");

            SuperController.singleton.onAtomUIDsChangedHandlers += OnAtomUIDsChanged;

            if (m_recognizer != null)
            {
                m_recognizer.Start();
            }

            RefreshKeywords();
        }

        public void OnDisable()
        {
            //SuperController.LogMessage(nameof(AAM_TrackerController) + ":" + "OnDisable");

            SuperController.singleton.onAtomUIDsChangedHandlers -= OnAtomUIDsChanged;

            if (m_recognizer != null)
            {
                m_recognizer.Stop();
            }
        }

        private void OnDestroy()
        {
            //SuperController.LogMessage(nameof(AAM_TrackerController) + ":" + "OnDestroy");

            if (m_recognizer != null)
            {
                m_recognizer.Dispose();
                m_recognizer = null;
            }
        }

        private JSONStorableStringChooser CreateTrackerIDChooser(string id, string label, string value)
        {
            JSONStorableStringChooser chooser = new JSONStorableStringChooser(id, m_tracker_ids, value, label);
            RegisterStringChooser(chooser);
            CreatePopup(chooser);

            return chooser;
        }

        //public UIDynamicTextField CreateTextInput(JSONStorableString jss, bool rightSide = false)
        //{
        //    var textfield = CreateTextField(jss, rightSide);
        //    textfield.height = 20f;
        //    textfield.backgroundColor = Color.white;
        //    var input = textfield.gameObject.AddComponent<InputField>();
        //    //var rect = input.GetComponent<RectTransform>().sizeDelta = new Vector2(1f, 0.4f);
        //    //input.lineType = InputField.LineType.SingleLine;
        //    input.textComponent = textfield.UItext;
        //    jss.inputField = input;
        //    return textfield;
        //}

        public override void Init()
        {
            pluginLabelJSON.val = PLUGIN_LABEL + " " + PLUGIN_VERSION;

            m_keywords_internal = new List<string>();
            m_keyword_to_action_map_internal = new Dictionary<string, List<Action>>();

            m_log_keyword_recognition = new JSONStorableBool("Log Keyword Recognition", true);
            RegisterBool(m_log_keyword_recognition); // Registering the value is important, so that it will be saved
            CreateToggle(m_log_keyword_recognition); // the second argument is for 'rightSide'. Supplying true will render the slider on the right-side

            m_tracker_ids = new List<string>();
            for (int i = 1; i < 9; i++)
            {
                m_tracker_ids.Add("Tracker" + i.ToString());
            }

            CreateSpacer(true).height = 50;

            // TODO: make an object for each line (label+tracker_chooser | keyword_text_input)
            // build callback handler for each text so that it updates the relevant keyword (or maybe just clear and rebuild entire keywords, like with other plugin's refreshing behavior)
            m_tracker_mappers.Add(new TrackerMapper(this, "Dildo", "<Auto> Dildo", "control", m_tracker_ids[0]));
            m_tracker_mappers.Add(new TrackerMapper(this, "His Pelvis", "<Auto> Male", "pelvisControl", m_tracker_ids[0]));
            m_tracker_mappers.Add(new TrackerMapper(this, "Her Head", "<Auto> Female", "headControl", m_tracker_ids[1]));
            m_tracker_mappers.Add(new TrackerMapper(this, "Her Pelvis", "<Auto> Female", "pelvisControl", m_tracker_ids[1]));

            CollectAtomsByGender();

            //OnTargetControllerUpdated(_targetControllerJSON.val);
            foreach (var tracker in m_tracker_mappers)
            {
                tracker.SyncDropDowns(tracker.m_atom_chooser.val);
            }
            //OnRigidbodyUpdated(_parentRigidbodyJSON.val);
        }

        protected void AddInternalKeywordAction(string keyword, Action action)
        {
            List<Action> actions;
            if (!m_keyword_to_action_map_internal.TryGetValue(keyword, out actions))
            {
                m_keywords_internal.Add(keyword);
                actions = new List<Action>();
                m_keyword_to_action_map_internal.Add(keyword, actions);
            }

            actions.Add(action);
        }

        // Runs once when plugin loads - after Init()
        protected void Start()
        {
            RefreshKeywords();
        }
    }
}
